import sre_constants
import logging
import locale
import csv
import re
from datetime import datetime
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from pybrowscap.loader import Browscap, TYPE_CSV


log = logging.getLogger(__name__)


# Url where latest version of csv browscap data file is located
URL = 'http://browsers.garykeith.com/stream.asp?BrowsCapCSV'


def load_file(browscap_file_path):
    """
    Loading browscap csv data file, parsing in into accessible python
    form and returning a new Browscap class instance with all appropriate data.

    :param browscap_file_path: location of browcap file on filesystem
    :type browscap_file_path: string
    :returns: Browscap instance filled with data
    :rtype: pybrowscap.loader.Browscap

    """
    def replace_defaults(line, defaults):
        """Replaces 'default' values for a line with parent line value and converting it into native python value.

        :param line: original line from browscap file
        :type line: dict
        :param defaults: default values for current line
        :type defaults: dict
        :returns: dictionary with replaced default values
        :rtype: dict
        :raises: IOError

        """
        new_line = {}
        for feature, value in line.iteritems():
            if value == 'default' or value == '':
                value = defaults[feature]
            if value == 'true':
                value = True
            if value == 'false':
                value = False
            if feature == 'MajorVersion' or feature == 'MinorVersion':
                try:
                    value = int(value)
                except Exception:
                    value = 0
            if feature == 'CSSVersion' or feature == 'AolVersion' or feature == 'Version':
                try:
                    value = float(value)
                except Exception:
                    value = float(0)
            new_line[feature.lower()] = value
        return new_line
    try:
        with open(browscap_file_path, 'rb') as csvfile:
            log.info('Reading browscap source file %s', browscap_file_path)
            dialect = csv.Sniffer().sniff(csvfile.read(4096))
            csvfile.seek(0)
            log.info('Getting file version and release date')
            csvfile.readline()
            line = csv.reader(StringIO(csvfile.readline())).next()
            log.info('Getting browcap file version')
            try:
                version = int(line[0])
            except ValueError:
                log.exception('Error while getting browscap file version')
                version = None
            log.info('Getting browscap file release date')
            try:
                old_locale = locale.getlocale()
                locale.setlocale(locale.LC_TIME, locale.normalize('en_US.utf8'))
                release_date = datetime.strptime(line[1][:-6], '%a, %d %b %Y %H:%M:%S')
            except (ValueError, locale.Error):
                log.exception('Error while getting browscap file release date')
                release_date = None
            finally:
                locale.setlocale(locale.LC_TIME, old_locale)

            log.info('Reading browscap user-agent data')
            reader = csv.DictReader(csvfile, dialect=dialect)
            defaults = {}
            browscap_data = {}
            regex_cache = []
            for line in reader:
                if line['Parent'] == 'DefaultProperties':
                    continue
                if '[{0}]'.format(line['Parent']) == line['UserAgent']:
                    defaults = line
                    continue
                line = replace_defaults(line, defaults)
                try:
                    ua_regex = '^{0}$'.format(re.escape(line['useragent'][1:-1]))
                    ua_regex = ua_regex.replace('\\?', '.').replace('\\*', '.*?')
                    browscap_data[ua_regex] = line
                    log.debug('Compiling user agent regex: %s', ua_regex)
                    regex_cache.append(re.compile(ua_regex))
                except sre_constants.error:
                    continue
        return Browscap(browscap_data, regex_cache, browscap_file_path, TYPE_CSV,
                        version, release_date)
    except IOError:
        log.exception('Error while reading browscap source file %s', browscap_file_path)
        raise


import unittest
import os

from pybrowscap.loader.csv import load_file


BROWSCAP = load_file(os.path.join(os.path.dirname(__file__), '..', '..', 'data', 'browscap_22_06_2011.csv'))


class TestBrowserFirefox(unittest.TestCase):

    user_agent = 'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2.18) Gecko/20110628 Ubuntu/10.10 (maverick) Firefox/3.6.18'

    def setUp(self):
        self.browser = BROWSCAP.search(self.user_agent)

    def tearDown(self):
        self.browser = None

    def test_items(self):
        self.assertDictEqual(self.browser.items(),
                             {'cookies': True, 'javaapplets': True, 'parent': 'Firefox 3.6',
                              'activexcontrols': False, 'backgroundsounds': False, 'win64': False,
                              'banned': False, 'javascript': True, 'syndicationreader': False,
                              'beta': False, 'sortorder': '860', 'aolversion': 0.0, 'alpha': False,
                              'frames': True, 'cssversion': 0.0, 'crawler': False,
                              'masterparent': False, 'tables': True, 'iframes': True,
                              'minorversion': 6, 'internalid': '11277', 'mobiledevice': False,
                              'vbscript': False, 'win32': False, 'platform': 'Linux', 'version': 3.6,
                              'useragent': '[Mozilla/5.0 (X11; *; *Linux*; *; rv:1.9.2*) Gecko/* Firefox/3.6*]',
                              'browser': 'Firefox', 'win16': False, 'majorversion': 3})

    def test_get(self):
        self.assertEqual(self.browser.get('platform'), 'Linux')
        self.assertEqual(self.browser.get('parent'), 'Firefox 3.6')
        self.assertIsNone(self.browser.get('codescale'))
        self.assertEqual(self.browser.get('codescale', ''), '')

    def test_name(self):
        self.assertEqual(self.browser.name(), 'Firefox')

    def test_category(self):
        self.assertEqual(self.browser.category(), 'Firefox 3.6')

    def test_platform(self):
        self.assertEqual(self.browser.platform(), 'Linux')

    def test_aol_version(self):
        self.assertIsInstance(self.browser.aol_version(), float)
        self.assertEqual(self.browser.aol_version(), 0.0)

    def test_version(self):
        self.assertIsInstance(self.browser.version(), float)
        self.assertEqual(self.browser.version(), 3.6)

    def test_version_major(self):
        self.assertIsInstance(self.browser.version_major(), int)
        self.assertEqual(self.browser.version_major(), 3)

    def test_version_minor(self):
        self.assertIsInstance(self.browser.version_minor(), int)
        self.assertEqual(self.browser.version_minor(), 6)

    def test_css_version(self):
        self.assertIsInstance(self.browser.css_version(), float)
        self.assertEqual(self.browser.css_version(), 0.0)

    def test_supports(self):
        self.assertTrue(self.browser.supports('tables'))

    def test_supports_tables(self):
        self.assertTrue(self.browser.supports_tables())

    def test_supports_frames(self):
        self.assertTrue(self.browser.supports_frames())

    def test_supports_iframes(self):
        self.assertTrue(self.browser.supports_iframes())

    def test_supports_java(self):
        self.assertTrue(self.browser.supports_java())

    def test_supports_javascript(self):
        self.assertTrue(self.browser.supports_javascript())

    def test_supports_vbscript(self):
        self.assertFalse(self.browser.supports_vbscript())

    def test_supports_activex(self):
        self.assertFalse(self.browser.supports_activex())

    def test_supports_cookies(self):
        self.assertTrue(self.browser.supports_cookies())

    def test_supports_css(self):
        self.assertFalse(self.browser.supports_css())

    def test_is_crawler(self):
        self.assertFalse(self.browser.is_crawler())

    def test_is_mobile(self):
        self.assertFalse(self.browser.is_mobile())

    def test_is_syndication_reader(self):
        self.assertFalse(self.browser.is_syndication_reader())

    def test_is_banned(self):
        self.assertFalse(self.browser.is_banned())

    def test_is_alpha(self):
        self.assertFalse(self.browser.is_alpha())

    def test_is_beta(self):
        self.assertFalse(self.browser.is_beta())

    def test_features(self):
        self.assertListEqual(self.browser.features(), ['tables', 'frames', 'iframes', 'javascript', 'cookies', 'java'])



class BrowserGooglebotTest(unittest.TestCase):

    user_agent = 'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)'

    def setUp(self):
        self.browser = BROWSCAP.search(self.user_agent)

    def tearDown(self):
        self.browser = None

    def test_items(self):
        self.assertDictEqual(self.browser.items(),
                             {'cookies': False, 'javaapplets': False, 'parent': 'Google',
                              'activexcontrols': False, 'backgroundsounds': False, 'win64': False,
                              'banned': False, 'javascript': False, 'syndicationreader': False,
                              'beta': False, 'sortorder': '100', 'aolversion': 0.0, 'alpha': False,
                              'frames': True, 'cssversion': 0.0, 'crawler': True, 'masterparent': False,
                              'tables': True, 'iframes': True, 'minorversion': 1,
                              'internalid': '4128', 'mobiledevice': False, 'vbscript': False,
                              'win32': False, 'platform': '', 'version': 2.1,
                              'useragent': '[*Googlebot/2.1*]', 'browser': 'Googlebot',
                              'win16': False, 'majorversion': 2})

    def test_get(self):
        self.assertEqual(self.browser.get('platform'), '')
        self.assertEqual(self.browser.get('parent'), 'Google')
        self.assertIsNone(self.browser.get('codescale'))
        self.assertEqual(self.browser.get('codescale', ''), '')

    def test_name(self):
        self.assertEqual(self.browser.name(), 'Googlebot')

    def test_category(self):
        self.assertEqual(self.browser.category(), 'Google')

    def test_platform(self):
        self.assertEqual(self.browser.platform(), '')

    def test_aol_version(self):
        self.assertIsInstance(self.browser.aol_version(), float)
        self.assertEqual(self.browser.aol_version(), 0)

    def test_version(self):
        self.assertIsInstance(self.browser.version(), float)
        self.assertEqual(self.browser.version(), 2.1)

    def test_version_major(self):
        self.assertIsInstance(self.browser.version_major(), int)
        self.assertEqual(self.browser.version_major(), 2)

    def test_version_minor(self):
        self.assertIsInstance(self.browser.version_minor(), int)
        self.assertEqual(self.browser.version_minor(), 1)

    def test_css_version(self):
        self.assertIsInstance(self.browser.css_version(), float)
        self.assertEqual(self.browser.css_version(), 0.0)

    def test_supports(self):
        self.assertTrue(self.browser.supports('tables'))

    def test_supports_tables(self):
        self.assertTrue(self.browser.supports_tables())

    def test_supports_frames(self):
        self.assertTrue(self.browser.supports_frames())

    def test_supports_iframes(self):
        self.assertTrue(self.browser.supports_iframes())

    def test_supports_java(self):
        self.assertFalse(self.browser.supports_java())

    def test_supports_javascript(self):
        self.assertFalse(self.browser.supports_javascript())

    def test_supports_vbscript(self):
        self.assertFalse(self.browser.supports_vbscript())

    def test_supports_activex(self):
        self.assertFalse(self.browser.supports_activex())

    def test_supports_cookies(self):
        self.assertFalse(self.browser.supports_cookies())

    def test_supports_css(self):
        self.assertFalse(self.browser.supports_css())

    def test_is_crawler(self):
        self.assertTrue(self.browser.is_crawler())

    def test_is_mobile(self):
        self.assertFalse(self.browser.is_mobile())

    def test_is_syndication_reader(self):
        self.assertFalse(self.browser.is_syndication_reader())

    def test_is_banned(self):
        self.assertFalse(self.browser.is_banned())

    def test_is_alpha(self):
        self.assertFalse(self.browser.is_alpha())

    def test_is_beta(self):
        self.assertFalse(self.browser.is_beta())

    def test_features(self):
        self.assertEqual(self.browser.features(), ['tables', 'frames', 'iframes'])

if __name__ == '__main__':
    unittest.main()


import os
import unittest
import urllib2

from pybrowscap.loader import Downloader


class DownloaderTest(unittest.TestCase):

    def setUp(self):
        self.url = 'http://www.codescale.net/en/'
        self.downloader = Downloader(self.url)
        self.new_file_path = os.path.join(os.path.dirname(__file__), '..', 'data', 'tmp'+'browscap_22_06_2011.csv')

    def tearDown(self):
        if os.path.exists(self.new_file_path):
            os.remove(self.new_file_path)

    def test_download_save_to_file(self):
        self.downloader.get(self.new_file_path)
        self.assertTrue(os.path.exists(self.new_file_path))
        self.assertTrue(os.path.isfile(self.new_file_path))

    def test_downloaded_file_content(self):
        self.downloader.get(self.new_file_path)
        with open(self.new_file_path, 'rb') as fp:
            contents = fp.read()
        self.assertNotEqual(contents.find('Software Development Done Right'), -1)

    def test_download_get_contents(self):
        contents = self.downloader.get()
        self.assertNotEqual(contents.find('Software Development Done Right'), -1)

    def test_download_no_url(self):
        with self.assertRaises(ValueError):
            self.downloader.url = ''
            self.downloader.get()

    def test_download_invalid_url(self):
        with self.assertRaises(urllib2.URLError):
            self.downloader.url = 'http://test'
            self.downloader.get()

    def test_download_invalid_proxy(self):
        with self.assertRaises(urllib2.URLError):
            self.downloader.proxy = 'http://test'
            self.downloader.get()

    def test_download_timeout_error(self):
        with self.assertRaises(urllib2.URLError):
            self.downloader.timeout = 0
            self.downloader.get()

    def test_download_http_error(self):
        with self.assertRaises(urllib2.HTTPError):
            self.downloader.url += 'a' * 4
            self.downloader.get()


if __name__ == '__main__':
    unittest.main()


import unittest
import os
from datetime import datetime

from pybrowscap.loader.csv import load_file
from pybrowscap.loader import Browscap, TYPE_CSV


class LoaderTest(unittest.TestCase):

    browscap_file1 = os.path.join(os.path.dirname(__file__), '..', '..', 'data', 'browscap_22_06_2011.csv')
    browscap_file2 = os.path.join(os.path.dirname(__file__), '..', '..', 'data', 'browscap_07_10_2011.csv')

    def test_load_csv_browscap(self):
        browscap = load_file(self.browscap_file1)
        self.assertIsInstance(browscap, Browscap)
        self.assertEqual(browscap.browscap_file_path, self.browscap_file1)
        self.assertEqual(browscap.type, TYPE_CSV)
        self.assertGreaterEqual(datetime.now(), browscap.loaded_at)
        self.assertIsNone(browscap.reloaded_at)
        self.assertEqual(len(browscap.data), 2)
        self.assertEqual(len(browscap.regex_cache), 2)
        self.assertEqual(browscap.version, 4856)
        self.assertEqual(browscap.release_date, datetime.strptime('Wed, 22 Jun 2011 23:26:51', '%a, %d %b %Y %H:%M:%S'))

    def test_reload_original_csv_browscap_file(self):
        browscap = load_file(self.browscap_file1)
        self.assertEqual(len(browscap.data), 2)
        self.assertEqual(len(browscap.regex_cache), 2)
        self.assertEqual(browscap.version, 4856)
        self.assertEqual(browscap.release_date, datetime.strptime('Wed, 22 Jun 2011 23:26:51', '%a, %d %b %Y %H:%M:%S'))
        browscap.reload()
        self.assertEqual(len(browscap.data), 2)
        self.assertEqual(len(browscap.regex_cache), 2)
        self.assertEqual(browscap.version, 4856)
        self.assertEqual(browscap.release_date, datetime.strptime('Wed, 22 Jun 2011 23:26:51', '%a, %d %b %Y %H:%M:%S'))

    def test_reload_new_csv_browscap_file(self):
        browscap = load_file(self.browscap_file1)
        self.assertEqual(len(browscap.data), 2)
        self.assertEqual(len(browscap.regex_cache), 2)
        self.assertEqual(browscap.version, 4856)
        self.assertEqual(browscap.release_date, datetime.strptime('Wed, 22 Jun 2011 23:26:51', '%a, %d %b %Y %H:%M:%S'))
        browscap.reload(self.browscap_file2)
        self.assertEqual(len(browscap.data), 2)
        self.assertEqual(len(browscap.regex_cache), 2)
        self.assertEqual(browscap.version, 4862)
        self.assertEqual(browscap.release_date, datetime.strptime('Fri, 07 Oct 2011 06:46:46', '%a, %d %b %Y %H:%M:%S'))


    def test_load_browscap_no_file(self):
        self.assertRaises(Exception, load_file, ('www.codescale.net'))

if __name__ == '__main__':
    unittest.main()


import logging
import urllib2
from datetime import datetime

from pybrowscap import Browser
load_csv_file = None


log = logging.getLogger(__name__)


TYPE_CSV = 1


class Downloader(object):
    """
    This class is responsible for downloading new versions of browscap file.
    It is possible to to define a timeout for connection and proxy server settings.

    """

    def __init__(self, url, timeout=20, proxy=None, additional_handlers=None):
        """Constructor

        Args:

        :param url: URL of browscap file
        :type url: string
        :param timeout: connection timeout
        :type timeout: int
        :param proxy: url of proxy server
        :type proxy: string
        :param additional_handlers: list of additional urllib2 handlers
        :type additional_handlers: list
        :returns: Downloader instance
        :rtype: Downloader

        """
        self.url = url
        self.timeout = timeout
        self.proxy = proxy
        self.additional_handlers = additional_handlers

    def get(self, save_to_file_path=None):
        """
        Getting browscap file contents and saving it to file or returning it as a string.
        Returns file contents if save_to_file_path is not None, file contents as string otherwise.

        ;param save_to_file_path: path on filesystem where browscap file will be saved
        :type save_to_file_path: string
        :returns: None or browscap file contents
        :rtype: string
        :raises: ValueError, urllib2.URLError, urllib2.HTTPError

        """
        try:
            log.info('Downloading latest version of browscap file from %s', self.url)
            opener = urllib2.build_opener()
            if self.proxy is not None:
                log.info('Setting up proxy server %s' % self.proxy)
                opener.add_handler(urllib2.ProxyHandler({'http': self.proxy}))
                if self.additional_handlers is not None:
                    for handler in self.additional_handlers:
                        opener.add_handler(handler)
            opener.addheaders = [('User-agent', 'pybrowscap downloader')]
            urllib2.install_opener(opener)
            response = opener.open(self.url, timeout=self.timeout)
            contents = response.read()
            response.close()
        except ValueError:
            log.exception('Url to browscap file is probably invalid')
            raise
        except urllib2.URLError:
            log.exception('Something went wrong while processing urllib2 handlers')
            raise
        except urllib2.HTTPError:
            log.exception('Something went wrong while downloading browscap file')
            raise

        if save_to_file_path is not None:
            try:
                log.info('Saving latest version of browscap file to %s', save_to_file_path)
                with open(save_to_file_path, 'wb') as file:
                    file.write(contents)
            except IOError:
                log.exception('Error while saving latest version of browscap file')
                raise
        else:
            return contents


class Browscap(object):
    """
    Browscap class represents abstraction on top of browscap data file.
    It contains all browscap data file data in python accessible form.

    """

    cache = {}

    def __init__(self, data_dict, regex_cache, browscap_file_path, type, version, release_date):
        """Constructor

        :param data_dict: dictionary of regex:line pairs
        :type data_dict: dict
        :param regex_cache: list of compiled regex patterns
        :type regex_cache: list
        :param browscap_file_path: path to the browscap file
        :type browscap_file_path: string
        :param type: type of browscap file csv, ini...
        :type type: int
        :param version: browscap file version
        :type version: int
        :param release_date: release date of browscap file
        :type release_date: datetime.datetime
        :returns: Browscap instance
        :rtype: pybrowscap.loader.Browscap

        """

        self.data = data_dict
        self.regex_cache = regex_cache
        self.browscap_file_path = browscap_file_path
        self.type = type
        self.version = version
        self.release_date = release_date
        self.loaded_at = datetime.now()
        self.reloaded_at = None

    def reload(self, browscap_file_path=None):
        """
        Reloads new data to this Browscap instance. This is useful
        mainly in apps that run in long living threads, like django projects.

        :param browscap_file_path: location of new browcap file on filesystem, or use old
        :type browscap_file_path: string or None
        :returns: None
        :raises: IOError

        """
        global load_csv_file
        if load_csv_file is None:
            from pybrowscap.loader.csv import load_file as load_csv_file
        try:
            file_to_load = browscap_file_path or self.browscap_file_path
            log.info('Reloading browscap instance with file %s', file_to_load)
            if self.type == TYPE_CSV:
                reloaded_browscap = load_csv_file(file_to_load)
            self.data = reloaded_browscap.data
            self.regex_cache = reloaded_browscap.regex_cache
            self.version = reloaded_browscap.version
            self.release_date = reloaded_browscap.release_date
            self.browscap_file_path = file_to_load
            self.cache = {}
            self.reloaded_at = datetime.now()
        except IOError:
            log.exception('Error while reloading Browscap instance with file %s', file_to_load)
            raise

    def search(self, user_agent_string):
        """Searching browscap data file for longest user_agent_string that matches the regex.

        :param user_agent_string: user agent to search in browscap file
        :type user_agent_string: string
        :returns: Browser instance or None
        :rtype: pybrowscap.Browser or None

        """
        log.info('Searching for user-agent: %s', user_agent_string)
        if user_agent_string in self.cache:
            log.debug('Cache-hit, searching skipped')
            return Browser(self.cache[user_agent_string])

        ua_regex_string = ''
        for ua_pattern in self.regex_cache:
            if ua_pattern.match(user_agent_string) and len(ua_pattern.pattern) > len(ua_regex_string):
                ua_regex_string = ua_pattern.pattern
        if ua_regex_string == '':
            log.info('No match found')
            return None
        else:
            log.info('Match found, returning Browser instance for: %s', ua_regex_string)
            self.cache[user_agent_string] = self.data[ua_regex_string]
            return Browser(self.data[ua_regex_string])